@@ -3,26 +3,36 @@ use warnings;
 use strict;
 use 5.010;
 
-use base 'App::Spec::Run';
+use Ref::Util qw/ is_arrayref is_hashref /;
+
+use base 'App::Spec::Run::Cmd';
 
 sub _dump_hash {
     my ($self, $hash) = @_;
     my @strings;
     for my $key (sort keys %$hash) {
-    next unless defined $hash->{ $key };
-        push @strings, "$key=$hash->{ $key }";
+        next unless defined $hash->{ $key };
+        my $value = $hash->{ $key };
+        if (is_hashref($value)) {
+            for my $key2 (sort keys %$value) {
+                push @strings, "$key=($key2=$value->{ $key2 })";
+            }
+        }
+        else {
+            push @strings, "$key=$value";
+        }
     }
     return join ",", @strings;
 }
 
 sub cook {
-    my ($self) = @_;
-    my $param = $self->parameters;
-    my $opt = $self->options;
+    my ($self, $run) = @_;
+    my $param = $run->parameters;
+    my $opt = $run->options;
     if ($ENV{PERL5_APPSPECRUN_TEST}) {
-        say "Subcommands: cook";
-        say "Options: " . $self->_dump_hash($opt);
-        say "Parameters: " .  $self->_dump_hash($param);
+        $run->out("Subcommands: cook");
+        $run->out("Options: " . $self->_dump_hash($opt));
+        $run->out("Parameters: " .  $self->_dump_hash($param));
         return;
     }
 
@@ -34,8 +44,8 @@ sub cook {
     push @with, $with if $with;
     push @with, "sugar" if $opt->{sugar};
 
-    say "Starting to cook $param->{drink}"
-    . (@with ? " with ". (join " and ", @with) : '');
+    $run->out("Starting to cook $param->{drink}"
+    . (@with ? " with ". (join " and ", @with) : ''));
 }
 
 my %countries = (
@@ -55,16 +65,16 @@ my %countries = (
 );
 
 sub weather {
-    my ($self) = @_;
-    my $param = $self->parameters;
+    my ($self, $run) = @_;
+    my $param = $run->parameters;
     my $cities = $param->{city};
     for my $city (@$cities) {
         my $info = $countries{ $param->{country} }->{ $city };
         my $output = sprintf "Weather in %s/%s: %s", $param->{country}, $city, $info->{weather};
-        if ($self->options->{temperature}) {
+        if ($run->options->{temperature}) {
             my $temp = $info->{temperature};
             my $label = "°C";
-            if ($self->options->{fahrenheit}) {
+            if ($run->options->{fahrenheit}) {
               $temp = int($temp * 9 / 5 + 32);
               $label = "F";
             }
@@ -79,19 +89,19 @@ sub countries {
 }
 
 sub cities {
-    my ($self) = @_;
-    my $country = $self->{options}->{country};
+    my ($self, $run) = @_;
+    my $country = $run->options->{country};
     my @countries = @$country ? @$country : sort keys %countries;
     say for map { sort keys %$_ } @countries{ @countries };
 }
 
 sub weather_complete {
-    my ($self, $args) = @_;
+    my ($self, $run, $args) = @_;
     my $runmode = $args->{runmode};
     return if $runmode ne "completion";
     my $comp_param = $args->{parameter};
 
-    my $param = $self->parameters;
+    my $param = $run->parameters;
     if ($comp_param eq "city") {
         my $country = $param->{country};
         my $cities = $countries{ $country } or return;
@@ -105,9 +115,15 @@ sub weather_complete {
 }
 
 sub palindrome{
-    my ($self) = @_;
-    my $string = $self->parameters->{string};
-    say +($string eq reverse $string) ? "yes" : "nope";
+    my ($self, $run) = @_;
+    my $string = $run->parameters->{string};
+    my $argv = $run->argv;
+    if (@$argv) {
+        $run->err("Sorry, only one palindrome at a time");
+        $run->halt(1);
+        return;
+    }
+    $run->out( ($string eq reverse $string) ? "yes" : "nope" );
 }
 
 my %units = (
@@ -164,8 +180,8 @@ my %conversions = (
 );
 
 sub convert {
-    my ($self) = @_;
-    my $param = $self->parameters;
+    my ($self, $run) = @_;
+    my $param = $run->parameters;
     my $type = $param->{type};
     my $source = $param->{source};
     my $targets = $param->{target};
@@ -175,29 +191,78 @@ sub convert {
         my $sub = $conversions{ $type }->{ $key };
         my $result = $sub->($value);
         my $label = $units{ $type }->{ $target }->{label};
-        say "$result$label";
+        $run->out("$result$label");
     }
 }
 
+sub config {
+    my ($self, $run, $args) = @_;
+    my $opt = $run->options;
+    my $param = $run->parameters;
+    if ($ENV{PERL5_APPSPECRUN_TEST}) {
+        $run->out("Options: " . $self->_dump_hash($opt));
+        return;
+    }
+    warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$opt], ['opt']);
+}
+
+sub data {
+    my ($self, $run) = @_;
+    my $opt = $run->options;
+    my $item = $opt->{item};
+
+    my $ref;
+    if ($item eq 'hash') {
+        $ref = {
+            foo => 23,
+        };
+    }
+    elsif ($item eq 'table') {
+        $ref = [
+            [qw/ FOO BAR /],
+            [ qw/ a b /],
+            [ qw/ c d /],
+        ];
+    }
+
+    # instead of print Data::Dumper->Dump($ref) or doing your own YAML or JSON
+    # encoding simply pass it to App::Spec
+    $run->out("Data '$item':");
+    $run->out($ref);
+}
 
 sub convert_complete {
-    my ($self, $args) = @_;
+    my ($self, $run, $args) = @_;
+    my $errors = $run->validation_errors;
     my $runmode = $args->{runmode};
     return if ($runmode ne "completion" and $runmode ne "validation");
     my $comp_param = $args->{parameter};
-    my $param = $self->parameters;
+    my $param = $run->parameters;
 
     if ($comp_param eq 'type') {
         return [sort keys %units];
     }
-    elsif ($comp_param eq 'source') {
+    if (delete $errors->{parameters}->{type}) {
+        $run->err("Invalid type\n");
+        return;
+    }
+    if ($comp_param eq 'source') {
         my $type = $param->{type};
         my $units = $units{ $type };
         return [map {
             +{ name => $_, description => $units->{ $_ }->{label} }
         } keys %$units];
     }
-    elsif ($comp_param eq 'target') {
+    if (delete $errors->{parameters}->{source}) {
+        $run->err("Invalid source\n");
+        return;
+    }
+    delete $errors->{parameters}->{target};
+    if (delete $errors->{parameters}->{value}) {
+        $run->err("Invalid value\n");
+        return;
+    }
+    if ($comp_param eq 'target') {
         my $type = $param->{type};
         my $source = $param->{source};
         my $value = $param->{value};
@@ -209,10 +274,15 @@ sub convert_complete {
             my $key = $source . '_' . $unit;
             my $sub = $conversions{ $type }->{ $key };
             my $result = $sub->($value);
-            push @result, {
-                name => $unit,
-                description => "$result$label",
-            };
+            if ($ENV{PERL5_APPSPECRUN_TEST}) {
+                push @result, $unit;
+            }
+            else {
+                push @result, {
+                    name => $unit,
+                    description => "$result$label",
+                };
+            }
         }
         return \@result;
     }
